/*
  ics-merge.cc

  Copyright (C) 2011 Taiji Yamada <taiji@aihara.co.jp>
 */
#define array_sizeof(v) (sizeof(v)/sizeof(v[0]))
#include <iostream>
#include <fstream>
#include <string>
#include <boost/lexical_cast.hpp>
#if defined(USE_BOOST_XPRESSIVE)
#include "boost_xpressive_utility.hpp"
using namespace boost::xpressive;
typedef string_sregex_map string_regex_map;
#else
#include "boost_regex_utility.hpp"
#endif
using namespace std;
using namespace boost;

string format_monthly_yearly(smatch const &what)
{
  int monthly_interval = lexical_cast<int>(what[1].str());
  if (monthly_interval % 12)
    return what[0].str();
  return "RRULE:FREQ=YEARLY;INTERVAL=" + lexical_cast<string>(monthly_interval / 12);
}

int usage(ostream &os, int argc, char *argv[])
{
  (void)argc;
  os << "usage:\t" << argv[0] << " [--ol2k|ical] [--] -|filename.ics [...]" << endl;
  return os == cerr ? 1 : 0;
}

int main(int argc, char *argv[])
{
  map<string, pair<string, string> > replace_string_map;
  struct {
    const char *key, *regex_str;
    const char *format;
  } replace_string_array[] = {
    { "prodid",		"^PRODID:[^\\r\\n]*$", "PRODID:-//Taiji Yamada <taiji@aihara.co.jp>//ics-merge//EN" },
    { "all_day",	"DT(START|END)(?:;[^:]*):(\\d+)T000000", "DT$1;VALUE=DATE:$2" },
    { "semicolon_eoi",	";(?!\\r\\n)$", "" },
  };
  for (size_t i=0; i<array_sizeof(replace_string_array); ++i)
    replace_string_map[replace_string_array[i].key] =
      make_pair(replace_string_array[i].regex_str, replace_string_array[i].format);

  typedef string (*format_action)(smatch const &);
  map<string, pair<string, format_action> > replace_action_map;
  struct {
    const char *key, *regex_str;
    format_action format;
  } replace_action_array[] = {
    { "monthly_yearly",	"^RRULE:FREQ=MONTHLY;INTERVAL=(\\d+)", format_monthly_yearly },
  };
  for (size_t i=0; i<array_sizeof(replace_action_array); ++i)
    replace_action_map[replace_action_array[i].key] =
      make_pair(replace_action_array[i].regex_str, replace_action_array[i].format);

  streambuf *cin_rdbuf = cin.rdbuf();
  vector<string> filenames;
  int a;
  for (a=1; a<argc; ++a)
    if (string(argv[a]) == "--" && ++a)
      break;
    else if (string(argv[a]) == "--ol2k")
      replace_string_map["all_day"] = make_pair("DT(START|END);VALUE=DATE:(\\d+)", "DT$1;TZID=Japan:$2T000000");
    else if (string(argv[a]) == "--ical")
      replace_string_map["all_day"] = make_pair("DT(START|END)(?:;[^:]*):(\\d+)T000000", "DT$1;VALUE=DATE:$2");
    else
      filenames.push_back(argv[a]);

  if (!filenames.size())
    return usage(cerr, argc, argv);

  string top_vcalendar_epilogue;
  map<string, int> prologue_state_count, epilogue_state_count;
  string_regex_map regex_map;
  for (size_t i=0; i<filenames.size(); ++i) {
    vector<string> state;
    ifstream ifs(filenames[i].c_str());
    cin.rdbuf((filenames[i] != "-") ? ifs.rdbuf() : cin_rdbuf);
    string sb;
    while (getline(cin, sb)) {
      if (sb.size() && sb[sb.size()-1] == '\r') sb.resize(sb.size()-1);
      while (!cin.eof()) {
	char c = cin.get();
	cin.putback(c);
	if (c == ' ') {
	  string cb;
	  getline(cin, cb);
	  if (cb.size() && cb[cb.size()-1] == '\r') cb.resize(cb.size()-1);
	  sb += "\r\n" + cb;
	}
	else
	  break;
      }
      string str(sb);
      struct {
	bool search_result;
	smatch match;
      } prologue, epilogue;
      prologue.search_result =
	regex_match(str, prologue.match, regex_map["^BEGIN:(VCALENDAR|VEVENT)$"]);
      epilogue.search_result =
	regex_match(str, epilogue.match, regex_map["^END:(VCALENDAR|VEVENT)$"]);
      if (prologue.search_result) {
	state.push_back(prologue.match[1].str());
	++prologue_state_count[state.back()];
      }

      for (map<string, pair<string, string> >::iterator it=replace_string_map.begin(); it!=replace_string_map.end(); ++it)
	sb = regex_replace(sb, regex_map[(*it).second.first], (*it).second.second);
      for (map<string, pair<string, format_action> >::iterator it=replace_action_map.begin(); it!=replace_action_map.end(); ++it)
	sb = regex_replace(sb, regex_map[(*it).second.first], (*it).second.second);
      if (state.back() == "VCALENDAR" && prologue_state_count[state.back()] == 1 && epilogue_state_count["VEVENT"] == 1)
	top_vcalendar_epilogue += sb + "\r\n";
      else if ((state.back() == "VCALENDAR" && prologue_state_count[state.back()] == 1) ||
	       state.back() != "VCALENDAR")
	cout << sb + "\r\n";

      if (epilogue.search_result) {
	++epilogue_state_count[state.back()];
	state.pop_back();
      }
    }
  }
  cout << top_vcalendar_epilogue;
  return 0;
}
