/*
  ics2tsv.cc

  Copyright (C) 2011 Taiji Yamada <taiji@aihara.co.jp>
 */
#include <iostream>
#include <fstream>
#include <string>
#include "date_time_numeric.hpp"
#include <boost/lexical_cast.hpp>
#include <boost/format.hpp>
#if defined(USE_BOOST_XPRESSIVE)
#include "boost_xpressive_utility.hpp"
using namespace boost::xpressive;
typedef string_sregex_map string_regex_map;
#else
#include "boost_regex_utility.hpp"
#endif
using namespace std;
using namespace boost;

int usage(ostream &os, int argc, char *argv[])
{
  (void)argc;
  os << "usage:\t" << argv[0] << " [--date|todo] [--head-] [--] -|filename.ics [...]" << endl;
  return os == cerr ? 1 : 0;
}

int main(int argc, char *argv[])
{
  enum tsv_type {
    tsv_date,
    tsv_todo,
  } tsv_kind = tsv_date;
  bool with_header(true);
  streambuf *cin_rdbuf = cin.rdbuf();
  vector<string> filenames;
  int a;
  for (a=1; a<argc; ++a)
    if (string(argv[a]) == "--" && ++a)
      break;
    else if (string(argv[a]) == "--date")
      tsv_kind = tsv_date;
    else if (string(argv[a]) == "--todo")
      tsv_kind = tsv_todo;
    else if (string(argv[a]) == "--head-")
      with_header = false;
    else
      filenames.push_back(argv[a]);

  if (!filenames.size())
    return usage(cerr, argc, argv);

  struct {
    string summary, description, location, categories, transp;
    posix_time::ptime dtstart, dtend;
    string rdate, rrule/*, exrule, related_to*/;
    vector<string> exdate;
  } vevent, vevent0;
  struct {
    string completed, summary, description;
    posix_time::ptime due;
  } vtodo, vtodo0;
  map<string, int> prologue_state_count, epilogue_state_count;
  string_regex_map regex_map;
  for (size_t i=0; i<filenames.size(); ++i) {
    vector<string> state;
    ifstream ifs(filenames[i].c_str());
    cin.rdbuf((filenames[i] != "-") ? ifs.rdbuf() : cin_rdbuf);
    string sb;
    while (getline(cin, sb)) {
      if (sb.size() && sb[sb.size()-1] == '\r') sb.resize(sb.size()-1);
      while (!cin.eof()) {
	char c = cin.get();
	cin.putback(c);
	if (c == ' ') {
	  string cb;
	  getline(cin, cb);
	  if (cb.size() && cb[cb.size()-1] == '\r') cb.resize(cb.size()-1);
	  sb += "\r\n" + cb;
	}
	else
	  break;
      }
      string str(sb);
      struct {
	bool search_result;
	smatch match;
      } prologue, epilogue;
      prologue.search_result =
	regex_match(str, prologue.match, regex_map["^BEGIN:(VCALENDAR|VEVENT|VALARM|VTODO)$"]);
      epilogue.search_result =
	regex_match(str, epilogue.match, regex_map["^END:(VCALENDAR|VEVENT|VALARM|VTODO)$"]);
      if (prologue.search_result) {
	state.push_back(prologue.match[1].str());
	++prologue_state_count[state.back()];
      }

      if (state.back() == "VCALENDAR" && prologue_state_count[state.back()] == 1 &&
	  epilogue_state_count[(tsv_kind == tsv_date) ? "VEVENT" : "VTODO"] == 1)
	;
      else if (state.back() == "VCALENDAR" && prologue_state_count[state.back()] == 1)
	;
      else if (tsv_kind == tsv_date && state.back() == "VEVENT") {
	smatch m;
	sb = regex_replace(sb, regex_map["\\t"], "\\\\t");
	sb = regex_replace(sb, regex_map["\\r\\n"],
#if defined(USE_BOOST_XPRESSIVE)
			   "\\r\\n"
#else
			   "\\\\r\\\\n"
#endif
			   );
	if (regex_match(sb, m, regex_map["^SUMMARY(.*)$"])) {
	  vevent.summary = m[1];
	}
	else if (regex_match(sb, m, regex_map["^DESCRIPTION(.*)$"])) {
	  vevent.description = m[1];
	}
	else if (regex_match(sb, m, regex_map["^LOCATION(.*)$"])) {
	  vevent.location = m[1];
	}
	else if (regex_match(sb, m, regex_map["^CATEGORIES(.*)$"])) {
	  vevent.categories = m[1];
	}
	else if (regex_match(sb, m, regex_map["^DTSTART;VALUE=DATE:(.*)$"])) {
	  vevent.dtstart = posix_time::ptime(gregorian::from_undelimited_string(m[1]));
	}
	else if (regex_match(sb, m, regex_map["^DTSTART[^:]*:(.*)$"])) {
	  vevent.dtstart = posix_time::from_iso_string(m[1]);
	}
	else if (regex_match(sb, m, regex_map["^DTEND;VALUE=DATE:(.*)$"])) {
	  vevent.dtend = posix_time::ptime(gregorian::from_undelimited_string(m[1]));
	}
	else if (regex_match(sb, m, regex_map["^DTEND[^:]*:(.*)$"])) {
	  vevent.dtend = posix_time::from_iso_string(m[1]);
	}
	else if (regex_match(sb, m, regex_map["^TRANSP(.*)$"])) {
	  vevent.transp = m[1];
	}
	else if (regex_match(sb, m, regex_map["^RRULE(.*)$"])) {
	  vevent.rrule = m[1];
	}
	else if (regex_match(sb, m, regex_map["^RDATE(.*)$"])) {
	  vevent.rdate = m[1];
	}
	else if (regex_match(sb, m, regex_map["^EXDATE(.*)$"])) {
	  vevent.exdate.push_back(m[1]);
	}
      }
      else if (tsv_kind == tsv_todo && state.back() == "VTODO") {
	smatch m;
	sb = regex_replace(sb, regex_map["\\t"], "\\\\t");
	sb = regex_replace(sb, regex_map["\\r\\n"],
#if defined(USE_BOOST_XPRESSIVE)
			   "\\r\\n"
#else
			   "\\\\r\\\\n"
#endif
			   );
	if (regex_match(sb, m, regex_map["^COMPLETED(.*)$"])) {
	  vtodo.completed = m[1];
	}
	else if (regex_match(sb, m, regex_map["^SUMMARY(.*)$"])) {
	  vtodo.summary = m[1];
	}
	else if (regex_match(sb, m, regex_map["^DESCRIPTION(.*)$"])) {
	  vtodo.description = m[1];
	}
	else if (regex_match(sb, m, regex_map["^DUE;VALUE=DATE:(.*)$"])) {
	  vtodo.due = posix_time::ptime(gregorian::from_undelimited_string(m[1]));
	}
	else if (regex_match(sb, m, regex_map["^DUE[^:]*:(.*)$"])) {
	  vtodo.due = posix_time::from_iso_string(m[1]);
	}
      }

      if (tsv_kind == tsv_date && epilogue.search_result && epilogue.match[1].str() == "VEVENT") {
	if (with_header && !epilogue_state_count["VEVENT"])
	  cout << "#date" << endl;
	if (vevent.dtstart.time_of_day() == posix_time::time_duration() && vevent.dtend.time_of_day() == posix_time::time_duration()) {
	  cout << format("%1%") % gregorian::to_iso_string(vevent.dtstart.date());
	  if (vevent.dtstart + gregorian::days(1) != vevent.dtend)
	    cout << format("-%1%") % gregorian::to_iso_string(vevent.dtend.date());
	}
	else {
	  cout << format("%1%") % posix_time::to_iso_string(vevent.dtstart);
	  if (vevent.dtstart.date() != vevent.dtend.date())
	    cout << format("-%1%") % posix_time::to_iso_string(vevent.dtend);
	  else if (vevent.dtstart.time_of_day() != vevent.dtend.time_of_day())
	    cout << format("-%1%") % posix_time::to_iso_string(vevent.dtend.time_of_day());
	}
	if (vevent.rdate.size() > 1)		cout << format("\trdate%1%")		% vevent.rdate;
	if (vevent.exdate.size()) {
	  cout << format("\texdate%1%") % vevent.exdate[0];
	  for (size_t i=1; i<vevent.exdate.size(); ++i)
	    cout << format("|%1%") % vevent.exdate[i];
	}
	if (vevent.rrule.size() > 1)		cout << format("\trrule%1%")		% vevent.rrule;
	if (vevent.summary.size() > 1)		cout << format("\tsummary%1%")		% vevent.summary;
	if (vevent.description.size() > 1)	cout << format("\tdescription%1%")	% vevent.description;
	if (vevent.location.size() > 1)		cout << format("\tlocation%1%")		% vevent.location;
	if (vevent.categories.size() > 1)	cout << format("\tcategories%1%")	% vevent.categories;
	if (vevent.transp.size() > 1)		cout << format("\ttransp%1%")		% vevent.transp;
	cout << endl;
	vevent = vevent0;
      }
      else if (tsv_kind == tsv_todo && epilogue.search_result && epilogue.match[1].str() == "VTODO") {
	if (with_header && !epilogue_state_count["VTODO"])
	  cout << "#todo" << endl;
	if (vtodo.due.time_of_day() == posix_time::time_duration())
	  cout << format("%1%") % gregorian::to_iso_string(vtodo.due.date());
	else
	  cout << format("%1%") % posix_time::to_iso_string(vtodo.due);
	if (vtodo.completed.size() > 1)		cout << format("\tcompleted%1%")	% vtodo.completed;
	if (vtodo.summary.size() > 1)		cout << format("\tsummary%1%")		% vtodo.summary;
	if (vtodo.description.size() > 1)	cout << format("\tdescription%1%")	% vtodo.description;
	cout << endl;
	vtodo = vtodo0;
      }

      if (epilogue.search_result) {
	++epilogue_state_count[state.back()];
	state.pop_back();
      }
    }
  }
  return 0;
}
